from typing import Any, Literal, Mapping, Union

from pyarrow import Schema
from typing_extensions import TypeAlias  # type: ignore

from ray_elasticsearch._compat import Document, Query


# Determine index and query typing based on Elasticsearch DSL availability.
if Document is not NotImplemented and Query is not NotImplemented:
    from ray_elasticsearch._compat import Document, Query

    IndexType: TypeAlias = Union[type[Document], str]  # type: ignore[no-redef]
    QueryType: TypeAlias = Union[Query, Mapping[str, Any]]  # type: ignore[no-redef]
    SchemaType: TypeAlias = Union[Schema, type[Document]]  # type: ignore[no-redef]
else:
    IndexType: TypeAlias = str  # type: ignore[no-redef,misc]
    QueryType: TypeAlias = Mapping[str, Any]  # type: ignore[no-redef,misc]
    SchemaType: TypeAlias = Schema  # type: ignore[no-redef,misc]


OpType: TypeAlias = Literal["index", "create", "update", "delete"]



from typing import cast, Optional, Iterable, Union
from warnings import warn

from pyarrow import (
    Schema,
    Field,
    DataType,
    schema,
    field,
    string,
    struct,
    list_,
    map_,
    float16,
    float32,
    float64,
    int8,
    int16,
    int32,
    int64,
    bool_,
    uint64,
    date32,
    date64,
    fixed_shape_tensor,
    unify_schemas,
)
from typing_extensions import TypeAlias, TypedDict, NotRequired  # type: ignore[import]

from ray_elasticsearch._compat import (
    Elasticsearch,
    Document,
    InnerDoc,
    Field as EsField,
    Object,
    Nested,
)

_META_FIELDS: Iterable[Field] = [
    field("_index", string()),
    field("_type", string()),
    field("_id", string()),
    field("_version", int64()),
    field("_seq_no", int64()),
    field("_primary_term", int64()),
    field("_score", float64(), nullable=True),
]
_META_SCHEMA = schema(fields=_META_FIELDS)


def complete_schema(
    base_schema: Schema,
    source_fields: Optional[Iterable[str]] = None,
    meta_fields: Optional[Iterable[str]] = None,
) -> Schema:
    source_field_names = (
        set(source_fields) if source_fields is not None else set(base_schema.names)
    )
    source_schema = schema(
        [
            base_schema.field(name)
            for name in source_field_names
            if name in base_schema.names
        ]
    )

    meta_field_names = (
        set(meta_fields) if meta_fields is not None else set(_META_SCHEMA.names)
    )
    meta_schema = schema(
        [
            _META_SCHEMA.field(name)
            for name in meta_field_names
            if name in _META_SCHEMA.names
        ]
    )

    return unify_schemas([source_schema, meta_schema])


class _PropertyDict(TypedDict):
    type: str
    properties: NotRequired["_PropertiesDict"]


_PropertiesDict: TypeAlias = dict[str, _PropertyDict]


def schema_from_document(document: Union[type[Document], type[InnerDoc]]) -> Schema:
    """
    Derive a PyArrow schema from an elasticsearch-dsl Document class.

    This works by inspecting the mapping that would be generated for the Document class.

    Known limitations:
    - In Elasticsearch, any field can also hold arrays of values. Unless explicitly defined as `multi=True` in the Elasticsearch DSL model, this function will assume a single value, not a list.
    - For date fields, especially custom date formats, the mapped schema does not guarantee successful parsing of the dates in PyArrow.
    """
    mapping = document._doc_type.mapping
    properties = mapping.properties.properties
    return schema(
        [
            field(
                name=name,
                type=_field_to_field_data_type(properties[name]),
                nullable=not properties[name]._required,
            )
            for name in mapping
        ]
    )


def schema_from_elasticsearch(
    elasticsearch: Elasticsearch,
    index: str,
) -> Schema:
    """
    Fetch the mapping for the given index from Elasticsearch and convert it to a PyArrow schema.

    Known limitations:
    - In Elasticsearch, any field can also hold arrays of values. Unless explicitly defined as `multi=True` in the Elasticsearch DSL model, this function will assume a single value, not a list.
    - For date fields, especially custom date formats, the mapped schema does not guarantee successful parsing of the dates in PyArrow.
    """
    mapping_response = elasticsearch.indices.get_mapping(index=index)
    if (
        index not in mapping_response
        or "mappings" not in mapping_response[index]
        or "properties" not in mapping_response[index]["mappings"]
    ):
        raise ValueError(f"No mapping found for index '{index}'.")
    return _properties_dict_to_schema(
        mapping=mapping_response[index]["mappings"]["properties"]
    )


def _properties_dict_to_schema(mapping: _PropertiesDict) -> Schema:
    """
    Determine a PyArrow `Schema` from an Elasticsearch mapping, given as a dictionary of properties or Elasticsearch DSL class.
    """
    return schema(
        [
            field(
                name=name,
                type=_property_to_field_data_type(prop),
            )
            for name, prop in mapping.items()
        ]
    )


def _field_to_field_data_type(
    field: EsField,
    force_single: bool = False,
) -> DataType:
    """
    Determine the PyArrow `DataType` for a given Elasticsearch DSL `Field` instance.
    """
    if isinstance(field, Nested):
        return list_(struct(schema_from_document(field._doc_class)))
    elif field._multi and not force_single:
        return list_(_field_to_field_data_type(field, force_single=True))
    elif isinstance(field, Object):
        return struct(schema_from_document(field._doc_class))
    else:
        return _property_to_field_data_type(field.to_dict())


def _property_to_field_data_type(prop: _PropertyDict) -> DataType:
    """
    Determine the PyArrow `DataType` for a given Elasticsearch property, given as a dictionary.
    """

    prop_dict = cast(dict, prop)
    if "type" not in prop_dict and "properties" in prop_dict:
        return struct(_properties_dict_to_schema(prop_dict["properties"]))

    type = prop_dict["type"]
    if type == "alias":
        # https://www.elastic.co/docs/reference/elasticsearch/mapping-reference/field-alias
        raise NotImplementedError("Alias fields are not supported yet.")
    elif type == "boolean":
        return bool_()
    elif type == "byte":
        return int8()
    elif type == "short":
        return int16()
    elif type == "integer":
        return int32()
    elif type == "long":
        return int64()
    elif type == "unsigned_long":
        return uint64()
    elif type == "half_float":
        return float16()
    elif type == "float":
        return float32()
    elif type == "double":
        return float64()
    elif type == "scaled_float":
        warn(
            "Elasticsearch scaled_float is mapped to float64, precision may be lost.",
            UserWarning,
            stacklevel=2,
        )
        return float64()
    elif type == "rank_feature":
        return float64()
    elif type == "text":
        return string()
    elif type == "keyword":
        return string()
    elif type == "binary":
        warn(
            "Elasticsearch binary fields are mapped to Base64-encoded strings.",
            UserWarning,
            stacklevel=2,
        )
        return string()
    elif type == "ip":
        return string()
    elif type == "search_as_you_type":
        return string()
    elif type == "version":
        return string()
    elif type == "date":
        if "format" in prop_dict and all(
            format
            in (
                "date",
                "basic_date",
                "basic_ordinal_date",
                "basic_week_date",
                "ordinal_date",
                "strict_date",
                "strict_ordinal_date",
                "strict_week_date",
                "strict_weekyear",
                "strict_weekyear_week",
                "strict_weekyear_week_day",
                "strict_year",
                "strict_year_month",
                "strict_year_month_day",
                "week_date",
                "weekyear",
                "weekyear_week",
                "weekyear_week_day",
                "year",
                "year_month",
                "year_month_day",
                "yyyy-MM-dd",
            )
            for format in prop_dict["format"].split("||")
        ):
            # If all formats are date-only formats, we can use date32.
            # https://www.elastic.co/docs/reference/elasticsearch/mapping-reference/mapping-date-format
            return date32()
        return date64()
    elif type == "dense_vector":
        if "dims" not in prop_dict:
            raise ValueError("Dense vector property must have dims.")
        element_type: DataType
        if "element_type" in prop_dict:
            if prop_dict["element_type"] == "float":
                element_type = float32()
            elif prop_dict["element_type"] == "byte":
                element_type = int8()
            elif prop_dict["element_type"] == "bit":
                element_type = bool_()
            else:
                raise NotImplementedError(
                    f"Dense vector element type '{prop_dict['element_type']}' is not supported."
                )
        else:
            element_type = float32()
        return fixed_shape_tensor(element_type, [prop_dict["dims"]])
    elif type == "nested":
        if "properties" not in prop_dict:
            raise ValueError("Object property must have properties.")
        return list_(struct(_properties_dict_to_schema(prop_dict["properties"])))
    elif type == "rank_features":
        return map_(string(), float64())
    else:
        raise NotImplementedError(f"Property type '{type}' is not supported yet.")



from typing import Any, Iterable, cast, Iterator, Union, Optional

from pyarrow import (
    schema,
    field,
    string,
    int64,
    float64,
    struct,
    list_,
    Schema,
    Field,
    DataType,
    StructType,
    ListType,
)
from pytest import skip

from ray_elasticsearch import ElasticsearchDatasource
from ray_elasticsearch._compat import Document, InnerDoc, Text, Object, Nested


def _sort_fields(fields: Iterable[Field]) -> Iterator[Field]:
    for f in sorted(fields, key=lambda f: f.name):
        yield _sort_field(f)


def _sort_field(field: Field) -> Field:
    return field.with_type(_sort_data_type(field.type))


def _sort_data_type(field: DataType) -> DataType:
    if isinstance(field, StructType):
        return struct(_sort_fields(field.fields))
    if isinstance(field, ListType):
        return list_(_sort_data_type(field.value_type))
    return field


def _sort_schema(s: Schema) -> Schema:
    fields = _sort_fields(s)
    metadata = cast(dict[Union[bytes, str], Union[bytes, str]], s.metadata)
    return schema(fields, metadata)


def _assert_schema_equals(a: Optional[Schema], b: Optional[Schema]) -> None:
    if a is None or b is None:
        assert a == b
        return
    assert _sort_schema(a) == _sort_schema(b)


def test_given_schema() -> Any:
    datasource = ElasticsearchDatasource(
        schema=schema(
            [
                field("field1", string(), nullable=False),
                field("field2", string(), nullable=True),
            ]
        ),
    )

    actual = datasource.schema()
    expected = schema(
        [
            field("field1", string(), nullable=False),
            field("field2", string(), nullable=True),
        ]
    )
    _assert_schema_equals(expected, actual)


def test_dsl_schema() -> Any:
    if Document is NotImplemented:
        skip("Elasticsearch DSL is not installed.")

    class _InnerDoc(InnerDoc):
        field1 = Text(required=True)
        field2 = Text(required=False, multi=True)

    class _Document(Document):
        field1 = Text(required=True)
        field2 = Text(required=False, multi=True)
        object1 = Object(_InnerDoc, required=True)
        nested1 = Nested(_InnerDoc, required=True)

    datasource = ElasticsearchDatasource(index=_Document)

    actual = datasource.schema()
    expected_fields: Iterable[Field] = [
        field("field1", string(), nullable=False),
        field("field2", list_(string()), nullable=True),
        field(
            "object1",
            struct(
                [
                    field("field1", string(), nullable=False),
                    field("field2", list_(string()), nullable=True),
                ]
            ),
            nullable=False,
        ),
        field(
            "nested1",
            list_(
                struct(
                    [
                        field("field1", string(), nullable=False),
                        field("field2", list_(string()), nullable=True),
                    ]
                )
            ),
            nullable=False,
        ),
        field("_index", string()),
        field("_type", string()),
        field("_id", string()),
        field("_version", int64()),
        field("_seq_no", int64()),
        field("_primary_term", int64()),
        field("_score", float64(), nullable=True),
    ]
    expected = schema(expected_fields)
    _assert_schema_equals(expected, actual)


def test_dsl_schema_with_source_fields() -> Any:
    if Document is NotImplemented:
        skip("Elasticsearch DSL is not installed.")

    class _InnerDoc(InnerDoc):
        field1 = Text(required=True)
        field2 = Text(required=False, multi=True)

    class _Document(Document):
        field1 = Text(required=True)
        field2 = Text(required=False, multi=True)
        object1 = Object(_InnerDoc, required=True)
        nested1 = Nested(_InnerDoc, required=True)

    datasource = ElasticsearchDatasource(
        index=_Document,
        source_fields=["field1", "object1", "nested1"],
    )

    actual = datasource.schema()
    expected_fields: Iterable[Field] = [
        field("field1", string(), nullable=False),
        field(
            "object1",
            struct(
                [
                    field("field1", string(), nullable=False),
                    field("field2", list_(string()), nullable=True),
                ]
            ),
            nullable=False,
        ),
        field(
            "nested1",
            list_(
                struct(
                    [
                        field("field1", string(), nullable=False),
                        field("field2", list_(string()), nullable=True),
                    ]
                )
            ),
            nullable=False,
        ),
        field("_index", string()),
        field("_type", string()),
        field("_id", string()),
        field("_version", int64()),
        field("_seq_no", int64()),
        field("_primary_term", int64()),
        field("_score", float64(), nullable=True),
    ]
    expected = schema(expected_fields)
    _assert_schema_equals(expected, actual)


def test_pydantic_schema() -> Any:
    try:
        from elasticsearch_pydantic import BaseDocument
    except ImportError:
        skip("Did not find elasticsearch-pydantic package.")

    class _Document(BaseDocument):
        field1: str
        field2: Optional[list[str]]

    datasource = ElasticsearchDatasource(index=_Document)

    actual = datasource.schema()
    expected_fields: Iterable[Field] = [
        field("field1", string(), nullable=False),
        field("field2", list_(string()), nullable=True),
        field("_index", string()),
        field("_type", string()),
        field("_id", string()),
        field("_version", int64()),
        field("_seq_no", int64()),
        field("_primary_term", int64()),
        field("_score", float64(), nullable=True),
    ]
    expected = schema(expected_fields)
    _assert_schema_equals(expected, actual)


def test_pydantic_schema_with_source_fields() -> Any:
    try:
        from elasticsearch_pydantic import BaseDocument
    except ImportError:
        skip("Did not find elasticsearch-pydantic package.")

    class _Document(BaseDocument):
        field1: str
        field2: Optional[list[str]]

    datasource = ElasticsearchDatasource(
        index=_Document,
        source_fields=["field1"],
    )

    actual = datasource.schema()
    expected_fields: Iterable[Field] = [
        field("field1", string(), nullable=False),
        field("_index", string()),
        field("_type", string()),
        field("_id", string()),
        field("_version", int64()),
        field("_seq_no", int64()),
        field("_primary_term", int64()),
        field("_score", float64(), nullable=True),
    ]
    expected = schema(expected_fields)
    _assert_schema_equals(expected, actual)



from functools import cached_property
from typing import (
    AbstractSet,
    Any,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Union,
    Dict,
)

from pandas import DataFrame
from pyarrow import Table
from ray.data import Datasink
from ray.data._internal.execution.interfaces import TaskContext
from ray.data.block import Block

from ray_elasticsearch._compat import (
    Elasticsearch,
    streaming_bulk,
)
from ray_elasticsearch._model import IndexType, OpType


class ElasticsearchDatasink(Datasink):
    _index: Optional[IndexType]
    _op_type: Optional[OpType]
    _chunk_size: int
    _source_fields: Optional[Iterable[str]]
    _meta_fields: Optional[Iterable[str]]
    _max_chunk_bytes: int
    _max_retries: int
    _initial_backoff: Union[float, int]
    _max_backoff: Union[float, int]
    _client_kwargs: dict[str, Any]

    def __init__(
        self,
        index: Optional[IndexType] = None,
        op_type: Optional[OpType] = None,
        chunk_size: int = 500,
        source_fields: Optional[Iterable[str]] = None,
        meta_fields: Optional[Iterable[str]] = None,
        max_chunk_bytes: int = 100 * 1024 * 1024,
        max_retries: int = 0,
        initial_backoff: Union[float, int] = 2,
        max_backoff: Union[float, int] = 600,
        **client_kwargs,
    ) -> None:
        super().__init__()
        self._index = index
        self._op_type = op_type
        self._chunk_size = chunk_size
        self._source_fields = source_fields
        self._meta_fields = meta_fields
        self._max_chunk_bytes = max_chunk_bytes
        self._max_retries = max_retries
        self._initial_backoff = initial_backoff
        self._max_backoff = max_backoff
        self._client_kwargs = client_kwargs

    @staticmethod
    def _iter_block_rows(block: Block) -> Iterator[Mapping[str, Any]]:
        if isinstance(block, Table):
            yield from block.to_pylist()
        elif isinstance(block, DataFrame):
            for _, row in block.iterrows():
                yield row.to_dict()
        else:
            raise RuntimeError(f"Unknown block type: {type(block)}")

    @cached_property
    def _elasticsearch(self) -> Elasticsearch:
        return Elasticsearch(**self._client_kwargs)

    @cached_property
    def _index_name(self) -> Optional[str]:
        if self._index is None:
            return None
        elif isinstance(self._index, str):
            return self._index
        else:
            return self._index()._get_index(required=True)  # type: ignore

    @cached_property
    def _source_field_set(self) -> Optional[AbstractSet[str]]:
        if self._source_fields is None:
            return None
        return set(self._source_fields)

    @cached_property
    def _meta_field_set(self) -> Optional[AbstractSet[str]]:
        if self._meta_fields is None:
            return None
        return set(self._meta_fields)

    def _transform_row(self, row: Mapping[str, Any]) -> Dict[str, Any]:
        meta = {key: value for key, value in row.items() if key.startswith("_")}
        if self._meta_field_set is not None:
            meta = {
                key: value
                for key, value in meta.items()
                if key.removeprefix("_") in self._meta_field_set
            }
        if self._index_name is not None:
            meta["_index"] = self._index_name
        if self._op_type is not None:
            meta["_op_type"] = self._op_type

        source = {key: value for key, value in row.items() if not key.startswith("_")}
        if self._source_field_set is not None:
            source = {
                key: value
                for key, value in source.items()
                if key in self._source_field_set
            }

        if self._op_type is None:
            return {"_source": source, **meta}
        elif self._op_type in ("index", "create"):
            return {"_source": source, **meta}
        elif self._op_type == "update":
            return {"doc": source, **meta}
        elif self._op_type == "delete":
            return meta
        else:
            return {"_source": source, **meta}

    def write(
        self,
        blocks: Iterable[Block],
        ctx: TaskContext,
    ) -> None:
        results = streaming_bulk(
            client=self._elasticsearch,
            actions=(
                self._transform_row(row)
                for block in blocks
                for row in self._iter_block_rows(block)
            ),
            chunk_size=self._chunk_size,
            max_chunk_bytes=self._max_chunk_bytes,
            raise_on_error=True,
            raise_on_exception=True,
            max_retries=self._max_retries,
            initial_backoff=self._initial_backoff,
            max_backoff=self._max_backoff,
        )
        for _ in results:
            pass

    def get_name(self) -> str:
        if self._index_name is not None:
            return f"Elasticsearch({self._index_name})"
        else:
            return "Elasticsearch"

    @property
    def supports_distributed_writes(self) -> bool:
        return True

    @property
    def num_rows_per_write(self) -> Optional[int]:
        return self._chunk_size



from os import environ
from pyarrow import field, schema, int32
from ray import init
from ray.data import read_datasource
from ray_elasticsearch import ElasticsearchDatasource

init()

source = ElasticsearchDatasource(
    index=environ["ELASTICSEARCH_INDEX"],
    hosts=environ["ELASTICSEARCH_HOST"],
    http_auth=(
        environ["ELASTICSEARCH_USERNAME"],
        environ["ELASTICSEARCH_PASSWORD"],
    ),
    schema=schema([
        field(name="value", type=int32(), nullable=False)
    ])
)

print(f"Num rows: {source.num_rows()}")
res = read_datasource(source, concurrency=100)\
    .sum("value")
print(f"Read complete. Sum: {res}")  # 49995000



from importlib.metadata import PackageNotFoundError, version

# Try to determine package version.
try:
    __version__ = version("ray-elasticsearch")
except PackageNotFoundError:
    pass

# Re-export names.
from ray_elasticsearch._model import (
    IndexType as IndexType,
    QueryType as QueryType,
    SchemaType as SchemaType,
    OpType as OpType,
)
from ray_elasticsearch._sink import ElasticsearchDatasink as ElasticsearchDatasink
from ray_elasticsearch._source import ElasticsearchDatasource as ElasticsearchDatasource
from ray_elasticsearch._utils import (
    unwrap_document as unwrap_document,
    unwrap_documents as unwrap_documents,
)



from importlib.metadata import metadata, PackageNotFoundError, version
from packaging.version import Version


def _is_installed(distribution_name: str) -> bool:
    try:
        metadata(distribution_name=distribution_name)
        return True
    except PackageNotFoundError:
        return False


def _is_version_at_least(distribution_name: str, at_least_version: str) -> bool:
    version_str = version(distribution_name=distribution_name)
    return Version(version_str) >= Version(at_least_version)


if _is_installed("elasticsearch8"):
    from elasticsearch8 import Elasticsearch as Elasticsearch  # type: ignore[no-redef]
    from elasticsearch8.helpers import streaming_bulk as streaming_bulk  # type: ignore[no-redef]
elif _is_installed("elasticsearch7"):
    from elasticsearch7 import Elasticsearch as Elasticsearch  # type: ignore[no-redef,assignment]
    from elasticsearch7.helpers import streaming_bulk as streaming_bulk  # type: ignore[no-redef,assignment]
elif _is_installed("elasticsearch"):
    if not _is_version_at_least("elasticsearch", "7.0.0"):
        raise ImportError("Elasticsearch version 7.0.0 or higher is required.")
    from elasticsearch import Elasticsearch as Elasticsearch  # type: ignore[no-redef,assignment]
    from elasticsearch.helpers import streaming_bulk as streaming_bulk  # type: ignore[no-redef,assignment]
else:
    raise ImportError("Elasticsearch is not installed.")

if _is_installed("elasticsearch8-dsl"):
    if _is_version_at_least("elasticsearch8-dsl", "8.12.0"):
        raise ImportError("Elasticsearch DSL version below 8.12.0 is required.")
    from elasticsearch8_dsl import (  # type: ignore[no-redef]
        Document as Document,
        InnerDoc as InnerDoc,
        Field as Field,
        Object as Object,
        Nested as Nested,
        Text as Text,
    )
    from elasticsearch8_dsl.query import Query as Query  # type: ignore[no-redef]
elif _is_installed("elasticsearch7-dsl"):
    from elasticsearch7_dsl import (  # type: ignore[no-redef,assignment]
        Document as Document,
        InnerDoc as InnerDoc,
        Field as Field,
        Object as Object,
        Nested as Nested,
        Text as Text,
    )
    from elasticsearch7_dsl.query import Query as Query  # type: ignore[no-redef,assignment]
elif _is_installed("elasticsearch-dsl"):
    if not _is_version_at_least("elasticsearch-dsl", "7.0.0"):
        raise ImportError("Elasticsearch DSL version 7.0.0 or higher is required.")
    if _is_version_at_least("elasticsearch-dsl", "8.12.0"):
        raise ImportError("Elasticsearch DSL version below 8.12.0 is required.")
    from elasticsearch_dsl import (  # type: ignore[no-redef,assignment]
        Document as Document,
        InnerDoc as InnerDoc,
        Field as Field,
        Object as Object,
        Nested as Nested,
        Text as Text,
    )
    from elasticsearch_dsl.query import Query as Query  # type: ignore[no-redef,assignment]
else:
    Document = NotImplemented  # type: ignore
    InnerDoc = NotImplemented  # type: ignore
    Field = NotImplemented  # type: ignore
    Object = NotImplemented  # type: ignore
    Nested = NotImplemented  # type: ignore
    Text = NotImplemented  # type: ignore
    Query = NotImplemented  # type: ignore



from typing import TypeVar, Callable, Any, Sequence

from pandas import DataFrame
from pyarrow import Table
from ray.data.block import DataBatch

from ray_elasticsearch._compat import Document


if Document is not NotImplemented:
    _D = TypeVar("_D", bound=Document)

    def _unwrap_document(
        doc_type: type[_D],
        row: dict[str, Any],
    ) -> _D:
        doc = {
            **{k: v for k, v in row.items() if k.startswith("_")},
            "_source": {k: v for k, v in row.items() if not k.startswith("_")},
        }
        return doc_type.from_es(doc)

    def unwrap_document(
        doc_type: type[_D],
    ) -> Callable[
        [Callable[[dict[str, Any], _D], dict[str, Any]]],
        Callable[[dict[str, Any]], dict[str, Any]],
    ]:
        """
        Convenience decorator for functions passed to `Dataset.map()` to directly use the decoded Elasticsearch DSL document.
        """

        def _decorate(
            map: Callable[[dict[str, Any], _D], dict[str, Any]],
        ) -> Callable[[dict[str, Any]], dict[str, Any]]:
            def _map(row: dict[str, Any]) -> dict[str, Any]:
                return map(row, _unwrap_document(doc_type, row))

            return _map

        return _decorate

    _B = TypeVar("_B", bound=DataBatch)

    def unwrap_documents(
        doc_type: type[_D],
    ) -> Callable[
        [Callable[[_B, Sequence[_D]], _B]],
        Callable[[_B], _B],
    ]:
        """
        Convenience decorator for functions passed to `Dataset.map_batches()` to directly use the sequence of decoded Elasticsearch DSL documents.
        """

        def _decorate(
            map: Callable[[_B, Sequence[_D]], _B],
        ) -> Callable[[_B], _B]:
            def _map(batch: _B) -> _B:
                df: DataFrame
                if isinstance(batch, Table):
                    df = batch.to_pandas()
                elif isinstance(batch, DataFrame):
                    df = batch
                else:
                    df = DataFrame(batch)
                documents: Sequence[_D] = [
                    _unwrap_document(doc_type, row.to_dict())
                    for _, row in df.iterrows()
                ]
                return map(batch, documents)

            return _map

        return _decorate
else:
    unwrap_document = NotImplemented  # type: ignore[assignment]
    unwrap_documents = NotImplemented  # type: ignore[assignment]



from typing import Any, Sequence

from pandas import DataFrame
from pytest import skip

from ray_elasticsearch import unwrap_document, unwrap_documents
from ray_elasticsearch._compat import Document, Text


def test_unwrap_document() -> None:
    if Document is NotImplemented:
        skip("Elasticsearch DSL is not installed.")

    class _Document(Document):
        field = Text(required=True)

    def _expected(row: dict[str, Any]) -> dict[str, Any]:
        return {**row, "custom": row["field"]}

    @unwrap_document(_Document)
    def _actual(row: dict[str, Any], document: _Document) -> dict[str, Any]:
        return {**row, "custom": document.field}

    data = _Document(field="example").to_dict(include_meta=True)
    data = {
        **{k: v for k, v in data.items() if k != "_source"},
        **data["_source"],
    }

    expected = _expected(data)
    actual = _actual(data)

    assert expected == actual


def test_unwrap_documents() -> None:
    if Document is NotImplemented:
        skip("Elasticsearch DSL is not installed.")

    class _Document(Document):
        field = Text(required=True)

    def _expected(batch: DataFrame) -> DataFrame:
        batch = batch.copy()
        batch["custom"] = [row["field"] for _, row in batch.iterrows()]
        return batch

    @unwrap_documents(_Document)
    def _actual(batch: DataFrame, documents: Sequence[_Document]) -> DataFrame:
        batch = batch.copy()
        batch["custom"] = [document.field for document in documents]
        return batch

    data = [_Document(field=f"example{i}").to_dict(include_meta=True) for i in range(5)]
    data = [
        {**{k: v for k, v in doc.items() if k != "_source"}, **doc["_source"]}
        for doc in data
    ]
    df = DataFrame(data)

    expected = _expected(df)
    actual = _actual(df)

    assert expected.equals(actual)



from os import environ
from ray import init
from ray.data import range
from ray_elasticsearch import ElasticsearchDatasink

init()

sink = ElasticsearchDatasink(
    index=environ["ELASTICSEARCH_INDEX"],
    hosts=environ["ELASTICSEARCH_HOST"],
    http_auth=(
        environ["ELASTICSEARCH_USERNAME"],
        environ["ELASTICSEARCH_PASSWORD"],
    ),
    source_fields=["value"],
    meta_fields=["id"],
)

range(10_000, concurrency=100)\
    .rename_columns({"id": "value"}, concurrency=100) \
    .add_column("_id", lambda df: df["value"], concurrency=100) \
    .write_datasink(sink, concurrency=100)
print("Write complete.")



from functools import cached_property
from typing import (
    AbstractSet,
    Any,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    TYPE_CHECKING,
)
from packaging.version import Version

from pyarrow import Schema, Table
from ray import __version__ as ray_version
from ray.data import Datasource, ReadTask
from ray.data.block import BlockMetadata

from ray_elasticsearch._compat import Elasticsearch, Query, Document
from ray_elasticsearch._schema import (
    complete_schema,
    schema_from_document,
    schema_from_elasticsearch,
)
from ray_elasticsearch._model import IndexType, QueryType, SchemaType


class ElasticsearchDatasource(Datasource):
    _index: Optional[IndexType]
    _query: Optional[QueryType]
    _keep_alive: str
    _chunk_size: int
    _source_fields: Optional[Iterable[str]]
    _meta_fields: Optional[Iterable[str]]
    _schema: Optional[SchemaType]
    _client_kwargs: dict[str, Any]

    def __init__(
        self,
        index: Optional[IndexType] = None,
        query: Optional[QueryType] = None,
        keep_alive: str = "5m",
        chunk_size: int = 1000,
        source_fields: Optional[Iterable[str]] = None,
        meta_fields: Optional[Iterable[str]] = None,
        schema: Optional[SchemaType] = None,
        **client_kwargs,
    ) -> None:
        super().__init__()
        self._index = index
        self._query = query
        self._keep_alive = keep_alive
        self._chunk_size = chunk_size
        self._source_fields = source_fields
        self._meta_fields = meta_fields
        self._schema = schema
        self._client_kwargs = client_kwargs

    @cached_property
    def _elasticsearch(self) -> Elasticsearch:
        return Elasticsearch(**self._client_kwargs)

    @cached_property
    def _index_name(self) -> Optional[str]:
        if self._index is None:
            return None
        elif isinstance(self._index, str):
            return self._index
        else:
            return self._index()._get_index(required=True)  # type: ignore

    @cached_property
    def _source_field_set(self) -> Optional[AbstractSet[str]]:
        if self._source_fields is None:
            return None
        return set(self._source_fields)

    @cached_property
    def _meta_field_set(self) -> Optional[AbstractSet[str]]:
        if self._meta_fields is None:
            return None
        return set(self._meta_fields)

    @cached_property
    def _query_dict(self) -> Optional[Mapping[str, Any]]:
        if self._query is None:
            return None
        elif Query is not NotImplemented and isinstance(self._query, Query):
            return self._query.to_dict()
        elif isinstance(self._query, dict):
            return self._query
        else:
            return None

    @cached_property
    def _safe_schema(self) -> Optional[Schema]:
        if self._schema is not None and isinstance(self._schema, Schema):
            return self._schema

        base_schema: Schema
        if (
            Document is not NotImplemented
            and self._schema is not None
            and isinstance(self._schema, type)
            and issubclass(self._schema, Document)
        ):
            base_schema = schema_from_document(
                document=self._schema,
            )
        if (
            Document is not NotImplemented
            and self._index is not None
            and isinstance(self._index, type)
            and issubclass(self._index, Document)
        ):
            base_schema = schema_from_document(
                document=self._index,
            )
        elif self._index_name is not None:
            base_schema = schema_from_elasticsearch(
                elasticsearch=self._elasticsearch,
                index=self._index_name,
            )
        else:
            return None

        return complete_schema(
            base_schema=base_schema,
            source_fields=self._source_fields,
            meta_fields=self._meta_fields,
        )

    def schema(self) -> Optional[Schema]:
        return self._safe_schema

    @cached_property
    def _num_rows(self) -> int:
        return self._elasticsearch.count(
            index=self._index_name,
            body=(
                {
                    "query": self._query_dict,
                }
                if self._query_dict is not None
                else {}
            ),
        )["count"]

    def num_rows(self) -> int:
        return self._num_rows

    @cached_property
    def _estimated_inmemory_data_size(self) -> Optional[int]:
        stats = self._elasticsearch.indices.stats(
            index=self._index_name,
            metric="store",
        )
        if "store" not in stats["_all"]["total"]:
            return None
        return stats["_all"]["total"]["store"]["total_data_set_size_in_bytes"]

    def estimate_inmemory_data_size(self) -> Optional[int]:
        return self._estimated_inmemory_data_size

    @staticmethod
    def _get_read_task(
        pit_id: str,
        query: Optional[Mapping[str, Any]],
        slice_id: int,
        slice_max: int,
        chunk_size: int,
        source_field_set: Optional[AbstractSet[str]],
        meta_field_set: Optional[AbstractSet[str]],
        schema: Optional[Schema],
        client_kwargs: dict,
    ) -> ReadTask:
        if Version(ray_version) >= Version("2.47.0") or TYPE_CHECKING:
            metadata = BlockMetadata(
                num_rows=None,
                size_bytes=None,
                input_files=None,
                exec_stats=None,
            )
        else:
            metadata = BlockMetadata(  # type: ignore[call-arg]
                num_rows=None,
                size_bytes=None,
                schema=schema,
                input_files=None,
                exec_stats=None,
            )

        def transform_row(row: Mapping[str, Any]) -> dict[str, Any]:
            meta: Mapping[str, Any] = {
                key: value
                for key, value in row.items()
                if key.startswith("_") and key != "_source"
            }
            if meta_field_set is not None:
                meta = {
                    key: value
                    for key, value in meta.items()
                    if key.removeprefix("_") in meta_field_set
                }

            source_dict: Mapping[str, Any] = row.get("_source", {})
            source: Mapping[str, Any] = {
                key: value for key, value in source_dict.items()
            }
            if source_field_set is not None:
                source = {
                    key: value
                    for key, value in source.items()
                    if key in source_field_set
                }

            return {
                **meta,
                **source,
            }

        def iter_blocks() -> Iterator[Table]:
            elasticsearch = Elasticsearch(**client_kwargs)
            search_after: Any = None
            while True:
                response = elasticsearch.search(
                    pit={"id": pit_id},
                    query=query,
                    slice={"id": slice_id, "max": slice_max},
                    size=chunk_size,
                    search_after=search_after,
                    sort=["_shard_doc"],
                    seq_no_primary_term=True,
                    version=True,
                )
                hits: Sequence[Mapping[str, Any]] = response["hits"]["hits"]
                if len(hits) == 0:
                    break
                search_after = max(hit["sort"] for hit in hits)
                rows: list[dict[str, Any]] = [transform_row(row) for row in hits]
                try:
                    yield Table.from_pylist(
                        mapping=rows,
                        schema=schema,
                    )
                except Exception as e:
                    print(
                        f"Failed to create Table with schema {schema}, from rows: {rows}."
                    )
                    raise e

        return ReadTask(
            read_fn=iter_blocks,
            metadata=metadata,
        )

    def get_read_tasks(
        self,
        parallelism: int,
        per_task_row_limit: Optional[int] = None,
    ) -> list[ReadTask]:
        pit_id: str = self._elasticsearch.open_point_in_time(
            index=self._index_name if self._index_name is not None else "_all",
            keep_alive=self._keep_alive,
        )["id"]
        try:
            return [
                self._get_read_task(
                    pit_id=pit_id,
                    query=self._query_dict,
                    slice_id=i,
                    slice_max=parallelism,
                    chunk_size=self._chunk_size,
                    source_field_set=self._source_field_set,
                    meta_field_set=self._meta_field_set,
                    client_kwargs=self._client_kwargs,
                    schema=self._safe_schema,
                )
                for i in range(parallelism)
            ]
        except Exception as e:
            self._elasticsearch.close_point_in_time(body={"id": pit_id})
            raise e

    def get_name(self) -> str:
        if self._index_name is not None:
            return f"Elasticsearch({self._index_name})"
        else:
            return "Elasticsearch"

    @property
    def supports_distributed_reads(self) -> bool:
        return True
